/*
 * NOTE(8080labs): This file inlines the file
 * https://github.com/naresh-n/slickgrid-column-data-autosize/blob/master/src/slick.autocolumnsize.js
 * The file is public under MIT license.
 * We do so to avoid too many dependencies.
 *
 * SlickGrid plugin: automatically resize a column by double clicking on the right edge
 * of a column header. You can also resize all columns by typing crtl+shift+a
 * This plugin has access to the slickgrid object it extends. We do that because we
 * need to reset the event handlers after we set column widths with grid.setColumns()
 */
(function($) {
    $.extend(true, window, {
        "Slick": {
            "AutoColumnSize": AutoColumnSize
        }
    });

    function AutoColumnSize(bamgrid) {
        const MIN_CELL_WIDTH = 65;
        const MAX_CELL_WIDTH = 300;
        const ADD_WIDTH_TO_HEADER = 15;

        var $container, context,
            keyCodes = {
                'A': 65
            };

        // AutoColumnSize is called by slick_grid.registerPlugin(), which passes AutoColumnSize
        // slick_grid. That's why we still have slick_grid here as an argument.
        function init(slick_grid) {
            // we do not use the slick_grid because there has been a weird bug
            // this has been fixed by always getting the current slickgrid via bamgrid
            $container = $(bamgrid.slick_grid.getContainerNode());
            $container.on("dblclick.autosize", ".slick-resizable-handle", reSizeColumn);
            $container.on("keydown", handleControlKeys);
            context = document.createElement("canvas").getContext("2d");
        }

        function destroy() {
            $container.off();
        }

        function handleControlKeys(event) {
            if (event.ctrlKey && event.shiftKey && event.keyCode === keyCodes.A) {
                resizeAllColumns(MIN_CELL_WIDTH);
            }
        }

        function resizeAllColumns(minWidth = 0) {
            var elHeaders = $container.find(".slick-header-column");
            var allColumns = bamgrid.slick_grid.getColumns();
            elHeaders.each(function(index, el) {
                var columnDef = $(el).data('column');
                var headerWidth = getElementWidth(el);
                var colIndex = bamgrid.slick_grid.getColumnIndex(columnDef.id);
                var column = allColumns[colIndex];
                // Add ADD_WIDTH_TO_HEADER to headers to compensate for the data type icon next to each
                // column header name
                var autoSizeWidth = Math.max(
                    minWidth,
                    headerWidth + ADD_WIDTH_TO_HEADER,
                    getMaxColumnTextWidth(columnDef, colIndex)
                );
                autoSizeWidth = Math.min(MAX_CELL_WIDTH, autoSizeWidth);
                column.width = autoSizeWidth;
            });
            bamgrid.slick_grid.setColumns(allColumns);
            bamgrid.makeSureResizingAColumnDoesntTriggerHeaderClickEvent();
            bamgrid.slick_grid.onColumnsResized.notify();
        }

        function reSizeColumn(e) {
            var headerEl = $(e.currentTarget).closest('.slick-header-column');
            var columnDef = headerEl.data('column');

            if (!columnDef || !columnDef.resizable) {
                return;
            }

            e.preventDefault();
            e.stopPropagation();

            var headerWidth = getElementWidth(headerEl[0]);
            var colIndex = bamgrid.slick_grid.getColumnIndex(columnDef.id);
            var allColumns = bamgrid.slick_grid.getColumns();
            var column = allColumns[colIndex];

            var autoSizeWidth = Math.max(headerWidth, getMaxColumnTextWidth(columnDef, colIndex)) + ADD_WIDTH_TO_HEADER;

            if (autoSizeWidth !== column.width) {
                column.width = autoSizeWidth;
                bamgrid.slick_grid.setColumns(allColumns);
                bamgrid.makeSureResizingAColumnDoesntTriggerHeaderClickEvent();
                bamgrid.slick_grid.onColumnsResized.notify();
            }
        }

        function getMaxColumnTextWidth(columnDef, colIndex) {
            var texts = [];
            var rowEl = createRow(columnDef);
            var grid_viewport = bamgrid.slick_grid.getViewport();
            var first_element_id = grid_viewport["top"];
            var last_element_id = grid_viewport["bottom"];

            for (var i = first_element_id; i <= last_element_id; i++) {
                texts.push(bamgrid.slick_grid.getDataItem(i)[columnDef.field]);
            }
            var template = getMaxTextTemplate(texts, columnDef, colIndex, rowEl);
            var width = getTemplateWidth(rowEl, template);
            deleteRow(rowEl);
            return width;
        }

        function getTemplateWidth(rowEl, template) {
            var cell = $(rowEl.find(".slick-cell"));
            cell.append(template);
            $(cell).find("*").css("position", "relative");
            return cell.outerWidth() + 5;
        }

        function getMaxTextTemplate(texts, columnDef, colIndex, rowEl) {
            var max = 0,
                maxTemplate = null;
            var formatFun = columnDef.formatter;
            $(texts).each(function(index, text) {
                var template;
                if (formatFun) {
                    template = $("<span>" + formatFun(index, colIndex, text, columnDef, data[index]) + "</span>");
                    text = template.text() || text;
                }
                var length = text ? getElementWidthUsingCanvas(rowEl, text) : 0;
                if (length > max) {
                    max = length;
                    maxTemplate = template || text;
                }
            });
            return maxTemplate;
        }

        function createRow(columnDef) {
            var rowEl = $('<div class="slick-row"><div class="slick-cell"></div></div>');
            rowEl.find(".slick-cell").css({
                "visibility": "hidden",
                "text-overflow": "initial",
                "white-space": "nowrap"
            });
            var gridCanvas = $container.find(".grid-canvas");
            $(gridCanvas).append(rowEl);
            return rowEl;
        }

        function deleteRow(rowEl) {
            $(rowEl).remove();
        }

        function getElementWidth(element) {
            var width, clone = element.cloneNode(true);
            clone.style.cssText = 'position: absolute; visibility: hidden;right: auto;text-overflow: initial;white-space: nowrap;';
            element.parentNode.insertBefore(clone, element);
            width = clone.offsetWidth;
            clone.parentNode.removeChild(clone);
            return width;
        }

        function getElementWidthUsingCanvas(element, text) {
            context.font = element.css("font-size") + " " + element.css("font-family");
            var metrics = context.measureText(text);
            return metrics.width;
        }

        return {
            init: init,
            destroy: destroy,
            resizeAllColumns: resizeAllColumns,
        };
    }
}(jQuery));
